# Decipherment with Baum-Welch training

## The problem

We observe a corpus of ciphertext $$c \in \mathcal{C}^{*}$$. We imagine that
there exists a corpus of plaintext $$p \in \mathcal{P}^{*}$$ which was used to
generate the ciphertext $$c$$. Our goal is to _decode_ the ciphertext; i.e., to
uncover the corresponding plaintext. We assume that the ciphertext $$c$$ has
been generated (i.e., encoded) by $$\pi_o\left(p \circ \gamma\right)$$ where
$$\pi_o$$ is the output projection operator and $$\gamma$$ is the encoder (i.e.,
inverse key). We further assume that the ciphertext can be recovered (i.e.,
decoded) using $$\pi_i\left(\gamma \circ c\right)$$ where $$\pi_i$$ is the input
projection operator. However, we do not observe either $$\gamma$$, and must
instead estimate it from data.

### Probabilistic formulation

We would like to estimate a conditional probability model which predicts the
plaintext $$p$$ given $$c$$; i.e., $$P(p | c)$$. By Bayes' rule, this is
equivalent to:

$$\begin{align}
P(p | c) &= \frac{P(p) P(c | p)}{P(c)} \\
         &\propto P(p) P(c | p)
\end{align}$$

That is; the conditional probability of the plaintext given observed ciphertext
is proportional to the product of the probability of the plaintext and the
conditional probability of the ciphertext given the plaintext; for any corpus of
ciphertext the denominator $$P(c)$$ is constant and thus can be ignored.

We then express decoding as:

$$\DeclareMathOperator*{\argmax}{arg\,max}
\hat{p} = \argmax_p P(p) P(c | p)$$

### Finite state formulation

We first construct a probabilistic model $$\Lambda$$ over $$\mathcal{P}^{*}$$,
which is a superset of the true plaintext $p$ and serves as a model of $$P(p)$$.
We henceforth refer to this as the _language model_.

We then construct a model of the inverse keyspace, $$ \Gamma : \mathcal{P}^{*}
\rightarrow \mathcal{C}^{*}$$. This is a superset of the true encoder relation
$$\gamma$$ and serves as an initial (usually uniform) model for $$P(c | p)$$; we
henceforth refer to it as the _channel model_.

Given an estimate for the channel model, $$\hat{\Gamma}$$, we can express
decoding as:

$$\hat{p} = \mathrm{ShortestPath}\left[\pi_i\left(\Lambda \circ \hat{\Gamma}
\circ c\right)\right]$$

where $$\textrm{ShortestPath}$$ represents the Viterbi algorithm.

### Channel model estimation

Knight et al. (2006) use _Baum-Welch estimation_ (a special case of _expectation
maximization_; Dempster et al. 1977) to estimate new weights for $$\Gamma$$ so
as to maximize $$P(p) P(c | p)$$.

During the _E-step_, we compute the cascade $$\Lambda \circ \hat{\Gamma} \circ
c$$. Each arc in the resulting cascade corresponds to an arc in
$$\hat{\Gamma}$$, and each arc in the cascade contributes an _expectation_ to
the corresponding $$\hat{\Gamma}$$ arc, computed as follows. We first use the
shortest distance algorithm to compute the forward costs $$\alpha$$ (i.e., the
cost of the shortest path from the start state to that state) and the backwards
costs $$\beta$$ (i.e., the cost of the shortest path from that state to a final
state) over this cascade. Then the expectation for a cascade arc with source
state $$s$$, weight $$W$$, and destination state $$d$$ is given by:

$$\alpha_s \otimes W \otimes \beta_d$$

where $$\otimes$$ is the semiring addition operator.

During the _M-step_, we simply sum (using the semiring plus operator $$\oplus$$)
all expectations in the cascade for each arc in $$\hat{\Gamma}$$, then normalize
these expectations (in a way to be made precise below) so that they define a
proper conditional probability distribution, then write these weights back into
$$\hat{\Gamma}$$. Thus the entire procedure consists of alternating the E- and
M-steps until convergence occurs, then decoding with the final $$\hat{\Gamma}$$.

## The library

### I/O

We adopt the following conventions in this library:

*   The ciphertext $$c$$ is expressed as as FAR of string FSTs
*   The language model $$\Lambda$$ is expressed as an FST
*   The channel model $$\Gamma$$ is expressed as an FST
*   The (hypothesized) plaintext $$p$$ is expressed as a FAR of string FSTs

### Training

The `fst::BaumWelch` template class is the primary driver for training. It
is templated on arc (which indirectly determines the strategy used for count
collection during the E-step; see below) as well as an expectation table type
(which determines the strategy used for renormalization during the M-step; see
below). It also exposes the matchers and composition filter template parameters
used to construct the cascade, though sensible default values are provided.

Most users should instead use the `fst::TrainBaumWelch` function. This
function takes as arguments FSTs representing the language model and channel
model and a FAR representing the ciphertext.

From the command line, users can use `fsttrainbaumwelch` with the
`--decipherment` flag.

#### Expectation tables and normalization strategies

This function also takes as an enum argument representing the desired
expectation table type. Three such types are available. The global expectation
table (`fst::GLOBAL`) performs global normalization and is only valid when
the channel model is itself acyclic. The state expectation table
(`fst::STATE`) normalizes so that all arcs leaving a state sum to semiring
one. The input-label expectation table (`fst::ILABEL`) assumes that the
channel model is a single-state machine and normalizes so that the weights of
all arcs leaving that single state with the same input label sum to semiring
one. The state/input-label expectation table (`fst::STATE_ILABEL`)
normalizes so that the weights of all arcs leaving a state with the same input
label sum to semiring one; this is arguably the most intuitive normalization
strategy and so it used as the default.

#### Semirings and count collection strategies

The training function is controlled in part by the arc type of the inputs.
In a
non-[_path_](http://www.openfst.org/twiki/bin/view/FST/FstWeightRequirements)
semiring like the log semiring, the E-step sums over all state sequences and
thus performs true Baum-Welch training. In a path semiring like the tropical
semiring, the E-step sums over only the most likely state sequence, an
approximation known as _Viterbi training_. In practice, Baum-Welch may be
marginally slower, though somewhat more accurate (Jansche 2003).

#### Other training options

Finally, `fst::TrainBaumWelch` takes an struct representing options for
training. These define the maximum number of iterations for training, whether or
not _random starts_ (to be defined) are to be performed, whether or not _zero
arcs_ (those arcs in the channel model which have zero expectations in the
cascade) are to be pruned beforehand, and a comparison/quantization delta to be
used to detect convergence.

Since expectation maximization training is only locally optimal, _random starts_
are used to avoid local minima by starting training from randomly weighted
channel models. For particularly difficult problems, it may be beneficial to
perform a very large number of random starts (e.g., Berg-Kilpatrick and Klein,
2013). In this library, we randomly initialize weights by sampling from the
uniform distribution (0, 1] in the real numbers, mapping these values onto the
appropriate values in the desired semiring, and then renormalizing.

#### Default training regime

The default training regime is as follows:

1.  Prune zero arcs
2.  Train the initial ("flat start") channel model till convergence or 50
    iterations, whichever comes first
3.  Generate 10 random starts and then train them until convergence or 50
    iterations, whichever comes first
4.  Return the converged model with the highest posterior likelihood.

To perform training with many random starts in parallel, use multiple
invocations of `fsttrainbaumwelch` with the `--decipherment` flag.

### Decoding

Decoding with the trained model is essentially a process of constructing a
weighted lattice followed by a shortest path computation.

In a non-[_path_](http://www.openfst.org/twiki/bin/view/FST/FstWeightRequirements) semiring like the log semiring, exact decoding is difficult to
perform in the absence of a deterministic lattice; however, determinizing the
entire lattice is rarely feasible for interesting problems. Therefore we use
the following strategy:

1. Compose the plaintext model, channel model, and ciphertext, and materialize
   the cascade, then input-project to produce an acyclic, epsilon-free
   non-deterministic weighted acceptor (henceforth, the NFA).
2. Compute $$\beta_n$$, the costs to the future in the NFA.
3. Compute, on the fly, a deterministic acceptor (henceforth, the DFA),
   Expansion of the DFA converts the NFA future costs $$\beta_n$$ to the DFA
   future costs $$\beta_d$$.
4. Compute, on the fly, a mapping of the DFA to a path semiring by converting
   the weights to tropical.
5. Compute the shortest path in the on-the-fly tropical-semiring DFA using 
   $$A^{*}$$ search (Hart et al. 1968) with $$\beta_d$$ as a heuristic, halting
   immediately after the shortest path is found.
6. Convert the shortest path back to the input semiring and remove weights.

In the case that this strategy is infeasible, one can simply convert the
three input FSTs to a path semiring like the tropical semiring ahead of time.

In a path semiring like the tropical semiring, exact decoding is much simpler:

1. Compose the plaintext model, channel model, and ciphertext, and materialize
   the cascade.
2. Compute the shortest path using the Viterbi algorithm.
3. Input-project, remove epsilons, and remove weights.

`fst::DecodeBaumWelch` (from C++) and `fstdecodebaumwelch` with the
`--decipherment` flag (from the command-line) can be used to invoke the above
procedures.

### Sparsity penalties

Knight et al. (2006) suggest an additional strategy for decoding. Observing that
trained channel model probabilities tend to be (impressionistically) too close
to uniform, they decode so as to maximize:

$$\DeclareMathOperator*{\argmax}{arg\,max}
\hat{p} = \argmax_p P(p) P(c | p)^k$$

where $$k$$ is some real number $$> 1$$ (Knight et al. suggest $$k = 3$$). This
strategy has the effect of imposing a simple sparsity bias because cubing (etc.)
and renormalizing the channel model probabilities increases the amount of the
probability mass assigned to likely events while reducing the probability mass
assigned to less likely events. `fst::DecodeBaumWelch` implements this
strategy using the `power` argument. (Note that this method does not
explicitly renormalize the channel model weights.)

## References

Berg-Kilpatrick, T., and Klein, D. 2013. Decipherment with a million random
restarts. In _Proc. EMNLP_, pages 874-878.

Dempster, A. P., Laird, N. M., and Rubin, D. B. 1977. Maximum likelihood from
incomplete data via the EM algorithm. _Journal of the Royal Statistical Society,
Series B_, 39(1): 1-38.

Hart, P. E., Nilsson, N. J., Raphael, B. 1968. A formal basis for the heuristic
determination of minimal cost paths. _IEEE Transactions on Systems Science and
Cybernetics_ 4(2): 100-107.

Jansche, M. 2003. Inference of string mappings for language technology. Doctoral
dissertation, Ohio State University.

Knight, K., Nair, A., Rashod, N., and Yamada, K. 2006. Unsupervised analysis for
decipherment problems. In _Proc. COLING_, pages 499-506.
